{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# VTA 测试 Demo"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "/media/pc/data/lxw/ai/tvm-book/doc/tutorials\n"
     ]
    }
   ],
   "source": [
    "%cd ..\n",
    "import set_env"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from PIL import Image\n",
    "import numpy as np\n",
    "import tvm\n",
    "from tvm import rpc, autotvm, relay\n",
    "from tvm.ir.transform import PassContext\n",
    "from tvm_book.transforms.common import FuseTransform\n",
    "from tvm_book.transforms import graphpack\n",
    "import vta\n",
    "from vta.testing import simulator\n",
    "# Make sure that TVM was compiled with RPC=1\n",
    "assert tvm.runtime.enabled(\"rpc\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "加载前端模型："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "from torch import nn\n",
    "\n",
    "class Model(nn.Module):\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.conv1 = nn.Conv2d(3, 50, 1, 1, 0, bias=False, groups=1)\n",
    "\n",
    "    def forward(self, x):\n",
    "        x = self.conv1(x)\n",
    "        # 下面两个等同\n",
    "        # x = x.view(x.size(0), -1)\n",
    "        x = torch.reshape(x, (x.size(0), -1))\n",
    "        # x = self.dense(x)\n",
    "        return x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# dev = tvm.cpu()\n",
    "# target = \"llvm\"\n",
    "input_shape = 1, 3, 8, 8\n",
    "input_data = torch.rand(input_shape).float()\n",
    "input_shapes = [(\"data\", input_shape)]\n",
    "model = Model().eval()\n",
    "trace_model = torch.jit.trace(model, [input_data.clone()])\n",
    "trace_model = trace_model.float().eval()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm.relay.testing import run_opt_pass\n",
    "import tvm\n",
    "from tvm import relay\n",
    "from tvm.relay import op\n",
    "from tvm.relay import ExprMutator\n",
    "from vta.top import graphpack"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "env = vta.get_env()\n",
    "with autotvm.tophub.context(env.target):\n",
    "    # 开始前端编译\n",
    "    mod, params = relay.frontend.from_pytorch(trace_model, input_shapes)\n",
    "    with PassContext(opt_level=3):\n",
    "        with relay.quantize.qconfig(global_scale=8.0,\n",
    "                                    skip_conv_layers=[]):\n",
    "            qmod = relay.quantize.quantize(mod, params=params)\n",
    "        qmod.show()\n",
    "        # 对 VTA target 进行 graph packing 和 constant folding\n",
    "        assert env.BLOCK_IN == env.BLOCK_OUT\n",
    "        # anf = run_opt_pass(mod[\"main\"], transform.ToANormalForm())\n",
    "        # anf = graphpack.get_subgraph(\n",
    "        #     mod[\"main\"], \n",
    "        #     start_name=\"cast\", \n",
    "        #     stop_name=\"cast\", \n",
    "        #     start_name_idx=None, \n",
    "        #     stop_name_idx=None, \n",
    "        #     count_meta = {}\n",
    "        # )\n",
    "        # print(anf)\n",
    "        # relay_prog = graphpack.graph_pack(\n",
    "        #     mod[\"main\"],\n",
    "        #     env.BATCH,\n",
    "        #     env.BLOCK_OUT,\n",
    "        #     env.WGT_WIDTH,\n",
    "        #     start_name=\"nn.conv2d\", #pack_dict[model][0],\n",
    "        #     stop_name=\"multiply\",\n",
    "        #     device_annot=(env.TARGET == \"intelfocl\"),\n",
    "        # )\n",
    "        # with vta.build_config(\n",
    "        #     opt_level=3,\n",
    "        #     disabled_pass={\"AlterOpLayout\",\n",
    "        #                     \"tir.CommonSubexprElimTIR\"}\n",
    "        # ):\n",
    "        #     lib = relay.build(relay_prog,\n",
    "        #                         target=env.target,\n",
    "        #                         params=params)\n",
    "        \n",
    "        # # # 将 inference library 发送到远程 RPC 服务器\n",
    "        # # lib.export_library(\"graphlib.tar\")\n",
    "        # # remote.upload(\"graphlib.tar\")\n",
    "        # # loaded_lib = remote.load_module(\"graphlib.tar\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm.relay.dataflow_pattern import is_op, wildcard, is_constant\n",
    "\n",
    "def make_preprocess_pattern():\n",
    "    r = is_op(\"multiply\")(wildcard(), is_constant())\n",
    "    r = is_op(\"round\")(r)\n",
    "    r = is_op(\"clip\")(r)\n",
    "    r = is_op(\"cast\")(r)\n",
    "    return r\n",
    "\n",
    "def make_conv2d_bias_pattern():\n",
    "    r = is_op(\"nn.conv2d\")(wildcard(), is_constant())\n",
    "    r = is_op(\"add\")(r, is_constant())\n",
    "    r = is_op(\"right_shift\")(r, is_constant())\n",
    "    r = is_op(\"clip\")(r)\n",
    "    r = is_op(\"cast\")(r)\n",
    "    return r\n",
    "\n",
    "# def make_output_pattern():\n",
    "#     r = is_op(\"cast\")(wildcard())\n",
    "#     r = is_op(\"multiply\")(r, wildcard())\n",
    "#     return r\n",
    "\n",
    "def make_reshape_output_pattern():\n",
    "    x = wildcard()\n",
    "    x = is_op(\"annotation.stop_fusion\")(x) | x\n",
    "    r = is_op(\"reshape\")(x)\n",
    "    r = is_op(\"cast\")(r)\n",
    "    r = is_op(\"multiply\")(r, is_constant())\n",
    "    return r"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "compiler_name = \"pack_special\"\n",
    "pattern_table = [\n",
    "    (f\"{compiler_name}.preprocess\", make_preprocess_pattern()),\n",
    "    (f\"{compiler_name}.conv2d_bias\", make_conv2d_bias_pattern()),\n",
    "    (f\"{compiler_name}.reshape_output\", make_reshape_output_pattern()),\n",
    "]\n",
    "merge_passes = tvm.transform.Sequential([\n",
    "    relay.transform.InferType(),\n",
    "    relay.transform.MergeComposite(pattern_table),\n",
    "    # # relay.transform.AnnotateTarget([compiler_name]),\n",
    "    # relay.transform.PartitionGraph(),\n",
    "    relay.transform.InferType(),\n",
    "    FuseTransform(),\n",
    "    relay.transform.InferType(),\n",
    "])\n",
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    run_qmod = merge_passes(qmod)\n",
    "run_qmod.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm_book.transforms import graphpack"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class ExprPack(ExprMutator):\n",
    "    \"\"\"Visitor to perform graph packing on an AST.\"\"\"\n",
    "\n",
    "    def __init__(self, bfactor, cfactor, weight_bits):\n",
    "        self.bfactor = bfactor\n",
    "        self.cfactor = cfactor\n",
    "        self.weight_bits = weight_bits\n",
    "        self.start_pack = False\n",
    "        # Cache Operator the algorithm matches against.\n",
    "        self.conv2d = op.op.get(\"nn.conv2d\")\n",
    "        self.conv2d_transpose = op.op.get(\"nn.conv2d_transpose\")\n",
    "        self.add = op.op.get(\"add\")\n",
    "        self.multiply = op.op.get(\"multiply\")\n",
    "        self.bias_add = op.op.get(\"nn.bias_add\")\n",
    "        self.pad = op.op.get(\"nn.pad\")\n",
    "        self.upsampling = op.op.get(\"nn.upsampling\")\n",
    "        self.reshape = op.op.get(\"reshape\")\n",
    "        self.number_of_conv2d = 0\n",
    "        self.unpack_transpose = True\n",
    "        super().__init__()\n",
    "\n",
    "    def visit_call(self, call):\n",
    "        oshape = graphpack._get_tensor_shape(call)\n",
    "        odtype = graphpack._get_tensor_type(call)\n",
    "        input_types = [arg.checked_type for arg in call.args]\n",
    "        args = [self.visit(arg) for arg in call.args]\n",
    "\n",
    "        if call.op == self.conv2d and odtype == \"int32\":\n",
    "            self.number_of_conv2d += 1\n",
    "            assert 8 % self.weight_bits == 0\n",
    "            w_lanes = 8 // self.weight_bits\n",
    "            data_layout = \"NCHW%dn%dc\" % (self.bfactor, self.cfactor)\n",
    "            kernel_layout = \"OIHW%do%di\" % (self.cfactor, self.cfactor)\n",
    "            data, weight = args\n",
    "            data_shape = graphpack._to_shape(input_types[0].shape)\n",
    "            kernel_shape = graphpack._to_shape(input_types[1].shape)\n",
    "            channels = call.attrs.channels\n",
    "            weight, kernel_shape, channels = graphpack._weight_shape_match(\n",
    "                weight, kernel_shape, channels, self.cfactor\n",
    "            )\n",
    "            kernel = graphpack._pack_weight(weight, kernel_shape, self.cfactor)\n",
    "            # insert bit packing when necessary\n",
    "            if w_lanes != 1:\n",
    "                assert 8 % w_lanes == 0\n",
    "                kernel = op.bitpack(kernel, lanes=w_lanes)\n",
    "\n",
    "            conv2d = op.nn.conv2d(\n",
    "                data,\n",
    "                kernel,\n",
    "                strides=call.attrs.strides,\n",
    "                padding=call.attrs.padding,\n",
    "                dilation=call.attrs.dilation,\n",
    "                groups=call.attrs.groups,\n",
    "                channels=channels,\n",
    "                kernel_size=call.attrs.kernel_size,\n",
    "                data_layout=data_layout,\n",
    "                kernel_layout=kernel_layout,\n",
    "                out_dtype=call.attrs.out_dtype,\n",
    "            )\n",
    "            return conv2d\n",
    "        return relay.Call(self.visit(call.op), args, call.attrs)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# bfactor = 1\n",
    "# cfactor = 16\n",
    "func = run_qmod[\"pack_special.conv2d_bias_1\"]\n",
    "# call = func.body\n",
    "# oshape = graphpack._get_tensor_shape(call)\n",
    "# odtype = graphpack._get_tensor_type(call)\n",
    "# input_types = [arg.checked_type for arg in call.args]\n",
    "transform = ExprPack(env.BATCH, env.BLOCK_OUT, env.WGT_WIDTH,)\n",
    "func = transform.visit(func)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "param = func.params[0]\n",
    "oshape = graphpack._get_tensor_shape(param)\n",
    "odtype = graphpack._get_tensor_type(param)\n",
    "# input_types = [arg.checked_type for arg in call.args]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "checked_type = param.checked_type"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "checked_type.shape"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "relay.var(param.name_hint, shape=oshape, dtype=odtype)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# binds = {\n",
    "#     for param in func.params\n",
    "# }\n",
    "# new_func = relay.bind(func, binds)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# transform = PreprocessPack(1, 16)\n",
    "# func = transform.visit(run_qmod[\"pack_special.preprocess_0\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# from tvm.contrib import graph_executor, download\n",
    "# # 下载 ImageNet categories\n",
    "# categ_url = \"https://github.com/uwsampl/web-data/raw/main/vta/models\"\n",
    "# categ_fn = \"synset.txt\"\n",
    "# download.download(f\"{categ_url}/{categ_fn}\", categ_fn)\n",
    "# synset = eval(open(categ_fn).read())\n",
    "# # 准备用于推理的测试图像\n",
    "# image = Image.open(\"tests/cat.jpg\").resize((32, 32))\n",
    "# # plt.imshow(image)\n",
    "# # plt.show()\n",
    "# image = np.array(image) - np.array([123.0, 117.0, 104.0])\n",
    "# image /= np.array([58.395, 57.12, 57.375])\n",
    "# image = image.transpose((2, 0, 1))\n",
    "# image = image[np.newaxis, :]\n",
    "# image = np.repeat(image, env.BATCH, axis=0)\n",
    "\n",
    "# with autotvm.tophub.context(env.target):\n",
    "#     # 生成图执行器（graph executor） `m`。\n",
    "#     m = graph_executor.GraphModule(lib[\"default\"](tvm.ext_dev(0)))\n",
    "#     # 设置网络参数和输入\n",
    "#     m.set_input(**params)\n",
    "#     m.set_input(\"data\", image)\n",
    "#     num = 4  # 为单个度量运行模块的次数\n",
    "#     rep = 3  # 测量的数量（由此得出 std dev）\n",
    "#     timer = m.module.time_evaluator(\"run\",\n",
    "#                                     tvm.ext_dev(0),\n",
    "#                                     number=num,\n",
    "#                                     repeat=rep)\n",
    "#     simulator.clear_stats()\n",
    "#     timer()\n",
    "#     sim_stats = simulator.stats()\n",
    "#     print(\"\\nExecution statistics:\")\n",
    "#     for k, v in sim_stats.items():\n",
    "#         # 由于多次执行 workload，需要 normalize 统计数据。\n",
    "#         # 注意，总是有一次 warm up 运行\n",
    "#         # 因此，将整体统计数据除以 (num * rep + 1)\n",
    "#         print(f\"\\t{k:<16}: {v // (num * rep + 1):>16}\")\n",
    "# tvm_output = m.get_output(0)\n",
    "# lib.ir_mod.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
